# -*- coding: WINDOWS-1251 -*-

# script      : BSP2CSDM
# revision    : 0.2.1
# date        : 2006/12/21
# author      : Simon Logic (slspam@land.ru)
# description : creates CSDM entity & spawn point files based on BSP entities data

""" You can customize the script algorithm by changing the following vars:
    
    * FL_EXPORT_ITEM_ANGLE
    * D_TRANSITION_RULES
    * LIST_SPAWN_CLASSES
    * FN_ENTITIES
    * FN_SPAWNS

    History:

    0.2.1 [2006/12/21]
    + added support for exporting item yaw angle (check if CSDM support this!)
    + var FL_EXPORT_ITEM_ANGLE (off by default)
    0.2.0 [2006/11/05]
    > improved speed (70x..1200x faster for the author's computer; actually
    depends on file size)
    + customizable vars LIST_SPAWN_CLASSES, FN_ENTITIES, FN_SPAWNS (see below)
    0.1.0 [2006/10/13]
    * first public release
"""
import os, sys, re, struct
from time import clock


FL_EXPORT_ITEM_ANGLE = 0 # flag: export item yaw angle

# here you can customize output file names (%s - is the name of the map without extension)...
FN_ENTITIES = "ents_%s.cfg"
#FN_SPAWNS = "csdm_%s.cfg"
FN_SPAWNS = "%s.spawns.cfg" # this one is for CSDM 2.x

# list of classes to convert into CSDM spawn points...
LIST_SPAWN_CLASSES = ("info_player_deathmatch",)
#LIST_SPAWN_CLASSES = ("info_player_deathmatch", "info_player_start", "info_vip_start")

""" D_TRANSITION_RULES manual:
    1) key names must be in lower case
    2) if the key value contains an empty string, then entity name
    is copied without replacing (e.g. 'item_battery')
    3) if you want to hide unsupported entity then just put a comment mark
    at the beginning of line or delete this line
"""
D_TRANSITION_RULES = { # HL DM entities only
    "weapon_9mmar":"weapon_mp5navy",        # machinegun with GL
    "weapon_crossbow":"weapon_awp",
    "weapon_rpg":"weapon_m249",             # RL
    "weapon_snark":"weapon_smokegrenade",   # snark is a hungry fantastic animal
    "weapon_handgrenade":"weapon_hegrenade",
    "weapon_gauss":"weapon_m4a1",           # lazer gun with recoil
    "weapon_hornetgun":"weapon_tmp",        # flies-gun
    "weapon_357":"weapon_deagle",           # magnum
    "weapon_satchel":"weapon_smokegrenade", # rcon mine
    "weapon_shotgun":"weapon_xm1014",                    
    "weapon_egon":"weapon_ak47",            # atomic disrupter
    "weapon_tripmine":"weapon_flashbang",
    "ammo_crossbow":"awp_ammo",
    "ammo_buckshot":"shotgun_ammo",         # shotgun ammo
    "ammo_gaussclip":"rifle_ammo",
    "ammo_rpgclip":"para_ammo",
    "ammo_9mmar":"smg_ammo",
    "ammo_argrenades":"weapon_hegrenade",
    "ammo_357":"pistol_ammo",
    "item_longjump":"",
    "item_battery":"",
    "item_healthkit":""
}

# SYSTEM VARS - DON'T CHANGE THEM!!!
SPAWN_TYPE = 0 # 0 - random spawn point
SF_LUMP = "2I" # bsp lump structure format
RE_ENTITY_BLOCK = re.compile(r'\{\s+(.*?classname.*?)^\s*\}', re.I|re.M|re.S)
RE_COUPLE_VALS = re.compile(r'"(.*?)"\s+"(.*?)"')
#------------------------------------------------------------------------------
def parseEntityBlock(_str):
    """ -> Dict
    """
    result = {}
    
    for str_pare in _str.split("\n"):
        mo = RE_COUPLE_VALS.search(str_pare)
        if mo:
            result[mo.group(1)] = mo.group(2)
    
    return result
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
if len(sys.argv) <> 2:
    print 
    print "Usage: python bsp2csdm.py <filename>"
    sys.exit(1)

if not os.path.exists(sys.argv[1]):
    print "File not found: ", sys.argv[1]

# prepare output filenames...
str_dir, str_basename = os.path.split(sys.argv[1])
str_basename = os.path.splitext(str_basename)[0]
str_fn_entities = os.path.join(str_dir, FN_ENTITIES % str_basename)
str_fn_spawns = os.path.join(str_dir, FN_SPAWNS % str_basename)

dict_skipped_classnames = {} # for debug & stats
list_error_entities = [] # e.g. classname presents in a rule, but 'origin' key not found

fo_ents = file(str_fn_entities, "w")
print "Created file:\n %s" % str_fn_entities
fo_spawns = file(str_fn_spawns, "w")
print "Created file:\n %s" % str_fn_spawns

print "Processing started..."
fo_bsp = file(sys.argv[1], "rb")
if 1:
    # parse the header and read entities data only
    fo_bsp.seek(4) # skip version
    list_data = struct.unpack(SF_LUMP, fo_bsp.read(struct.calcsize(SF_LUMP)))
    fo_bsp.seek(list_data[0])
    str_chunk = fo_bsp.read(list_data[1])
    list_data = None
else:
    # old method: slow but more reliable
    str_chunk = fo_bsp.read()
fo_bsp.close()
print "Read %d bytes in memory" % len(str_chunk)

clk_start = clock()
for str_entity in RE_ENTITY_BLOCK.findall(str_chunk):
     str_entity = str_entity.strip().lower()
     dict_entity = parseEntityBlock(str_entity)
     
     if dict_entity:
        str_classname = dict_entity["classname"]
        
        if D_TRANSITION_RULES.has_key(str_classname):
            if not dict_entity.has_key("origin"):
                list_error_entities.append(dict_entity)
            else:
                str_new_class = D_TRANSITION_RULES[str_classname]
                if not str_new_class:
                    str_new_class = str_classname
                if FL_EXPORT_ITEM_ANGLE and dict_entity.has_key("angle"):
                    fo_ents.write("%s %s %s\n" % (str_new_class, dict_entity["origin"], dict_entity["angle"]))
                else:
                    fo_ents.write("%s %s\n" % (str_new_class, dict_entity["origin"]))

        elif str_classname in LIST_SPAWN_CLASSES:
            if not dict_entity.has_key("origin"):
                list_error_entities.append(dict_entity)
            else:
                if dict_entity.has_key("angle"):
                    i_angle = int(dict_entity["angle"])
                else:
                    i_angle = 0
                # CSDM spawn point format: IDVec[3], vAngle[3], iTeam, oAngle[3]
                fo_spawns.write("%s 0 %d 0 %d 0 %d 0\n" % (dict_entity["origin"], i_angle, SPAWN_TYPE, i_angle))

        elif dict_skipped_classnames.has_key(str_classname):
            dict_skipped_classnames[str_classname] += 1
        
        else:
            dict_skipped_classnames[str_classname] = 1

clk_stop = clock()

fo_ents.close()
fo_spawns.close()

print "Processing finished in %.02f sec" % (clk_stop - clk_start)
print "Error entities: %d" % len(list_error_entities)
print "Skipped classnames: %d" % len(dict_skipped_classnames)
